from __future__ import annotations

from functools import lru_cache

from sqla_ext.inspect._to_core_table import to_core_table
from sqla_ext.protocols import TableCoercable


@lru_cache()
def to_table_name(entity: TableCoercable) -> str:
    r"""Get the name of a sqlalchemy table-like entity

    :param entity: A sqlalchemy :class:`Table`, :class:`DeclarativeBase` or :class:`Mapper`

    :return: :class:`str`
    """
    return str(to_core_table(entity).name)



from sqla_ext.inspect._to_core_table import to_core_table
from sqla_ext.inspect._to_schema_name import to_schema_name
from sqla_ext.inspect._to_table_name import to_table_name

__all__ = ["to_core_table", "to_table_name", "to_schema_name"]



from typing import Any, Dict

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.functions import FunctionElement
from sqlalchemy.types import DateTime


class utc_now(FunctionElement):
    r"""Current timestamp in UTC timezone

    :Dialects:
        - mysql
        - postgresql
        - sqlite

    :return: :class:`FuntionElement`

    E.g.::

        from sqlalchemy import select
        from sqla_ext import func as func_ext

        query = select([
            func_ext.datetime.utc_now()
        ])

    The above statement will produce SQL resembling::

        SELECT
            timezone('utc', current_timestamp)
    """

    name = "to_array"

    def __init__(self) -> None:
        super().__init__()

    type = DateTime()


@compiles(utc_now, "postgresql")
def pg(element: utc_now, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:
    return "timezone('utc', current_timestamp)"


@compiles(utc_now, "sqlite")
def sqlite(element: utc_now, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:
    return "datetime('now')"


@compiles(utc_now, "mysql")
def mysql(element: utc_now, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:
    return "utc_timestamp()"



from __future__ import annotations

from functools import lru_cache
from typing import Optional

from sqla_ext.inspect._to_core_table import to_core_table
from sqla_ext.protocols import TableCoercable


@lru_cache()
def to_schema_name(entity: TableCoercable) -> Optional[str]:
    r"""Get the schema of a sqlalchemy table-like entity

    :param entity: A sqlalchemy :class:`Table`, :class:`DeclarativeBase` or :class:`Mapper`

    :return: :class:`Optional[str]`
    """

    schema_name = to_core_table(entity).schema
    return str(schema_name) if schema_name is not None else None



from typing import Any, Dict, Union

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.elements import ColumnClause, TextClause
from sqlalchemy.sql.functions import GenericFunction
from sqlalchemy.types import JSON


class agg(GenericFunction):  # type: ignore
    r"""JSON array aggregation

    :Dialects:
        - mysql
        - postgresql
        - sqlite

    :param expression: A SQL expression, such as a
        :class:`ColumnElement` expression or a :class:`TextClause` with elements
        that are castable as JSON

    :return: :class:`FuntionElement`

    E.g.::

        from sqlalchemy import select, Integer, text, table, column
        from sqla_ext import func as func_ext

        t = table('some_table', column('q', Integer))

        query = select([
            func_ext.json.agg(t.q)
        ])

    The above statement will produce SQL resembling::

        SELECT
            jsonb_agg(some_table.q)
        FROM
            some_table
    """

    type = JSON()
    name = "agg"

    clauses: Any

    def __init__(self, expression: Union[ColumnClause, TextClause]) -> None:
        super().__init__(expression)


@compiles(agg, "postgresql")
def pg(element: agg, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:
    return "jsonb_agg(%s)" % compiler.process(element.clauses, **kw)


@compiles(agg, "sqlite")
def sqlite(element: agg, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:
    return "json_group_array(%s)" % compiler.process(element.clauses, **kw)


@compiles(agg, "mysql")
def mysql(element: agg, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:
    return "json_arrayagg(%s)" % compiler.process(element.clauses, **kw)



# pylint: disable= unsubscriptable-object
from __future__ import annotations

from typing import Any, Dict, Union

from sqlalchemy import func as sqla_func
from sqlalchemy import select
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import cast
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.elements import ColumnClause, TextClause
from sqlalchemy.sql.functions import GenericFunction
from sqlalchemy.sql.type_api import TypeEngine


class to_array(GenericFunction):  # type: ignore
    r"""Cast a json array to a native array

    :Dialects:
        - postgresql

    :param expression: A SQL expression, such as a
        :class:`ColumnElement` expression or a :class:`TextClause` that can be coerced into a JSONB array

    :param type\_: A :class:`.TypeEngine` class or instance indicating
        the type to ``CAST`` elements of the native array as.

    :return: :class:`FuntionElement`

    E.g.::

        from sqlalchemy import select, Integer, text
        from sqla_ext import func as func_ext

        query = select([
            func_ext.json.to_array(text("'[1,2,3,4]'::jsonb"), Integer)
        ])

    The above statement will produce SQL resembling::

        SELECT
            array_agg(CAST(anon_1 AS INTEGER)) AS array_agg_1
        FROM
            jsonb_array_elements(CAST('[1,2,3,4]'::jsonb AS JSONB)) AS anon_1
    """

    name = "to_array"

    def __init__(
        self, expression: Union[ColumnClause, TextClause], type_: TypeEngine
    ) -> None:
        super().__init__(expression, type_)


function = to_array


@compiles(function, "postgresql")
def pg(element: function, compiler: SQLCompiler, **kw: Dict[str, Any]) -> str:

    args = iter(element.clauses)  # type: ignore
    json_field = next(args)  # type: ignore
    type_bind_param = next(args)  # type: ignore
    type_: TypeEngine = type_bind_param.value  # type: ignore

    assert isinstance(type_, TypeEngine) or issubclass(type_, TypeEngine)

    select_from = sqla_func.jsonb_array_elements(cast(json_field, JSONB)).table_valued(
        "value"
    )
    statement = select([sqla_func.array_agg(cast(select_from.column, type_))])

    return compiler.process(statement, **kw)



from typing import Callable

import sqlalchemy.types as types
from sqlalchemy.dialects.postgresql.base import PGDialect, ischema_names


class CITEXT(types.Concatenable, types.UserDefinedType):
    r"""Case Insensitive Text Type

    :Dialects:
        - postgresql

    https://www.postgresql.org/docs/current/citext.html

    E.g.::

        from sqlalchemy import Column, Integer
        from sqlalchemy.orm import declarative_base
        from sqla_ext.types.postgresql import CITEXT


        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(Integer, primary_key=True)
            email = sa.Column(CITEXT())


        user = User(id=1, email='John.Smith@example.com')
        session.add(user)
        session.commit()

        # Note: query email is lowercase
        user = (
            session.query(User)
            .filter(User.email == 'john.smith@example.com')
            .one()
        )
        assert user.id == 1
    """

    def literal_processor(self, dialect: PGDialect) -> Callable:
        def process(value: str) -> str:
            value = value.replace("'", "''")

            if dialect.identifier_preparer._double_percents:
                value = value.replace("%", "%%")

            return "'%s'" % value

        return process

    def get_col_spec(self) -> str:
        return "CITEXT"

    def bind_processor(self, dialect: PGDialect) -> Callable:
        def process(value: str) -> str:
            return value

        return process

    def result_processor(self, dialect: PGDialect, coltype: int) -> Callable:
        def process(value: str) -> str:
            return value

        return process


# Register CIText to SQLAlchemy's Postgres reflection subsystem.
ischema_names["citext"] = CITEXT



from __future__ import annotations

from functools import lru_cache

from sqlalchemy import Table
from sqlalchemy.orm import Mapper

from sqla_ext.protocols import ORMTableProtocol, TableCoercable


@lru_cache()
def to_core_table(entity: TableCoercable) -> Table:
    r"""Coerces multiple SQLA Table-like entities to a core sqlalcehmy.Table

    :param entity: A sqlalchemy :class:`Table`, :class:`DeclarativeBase` or :class:`Mapper`

    :return: :class:`Table`
    """

    if isinstance(entity, Table):
        return entity

    if isinstance(entity, ORMTableProtocol):
        return entity.__table__

    if isinstance(entity, Mapper):
        # re-wrapping with to_core_table is not necessary
        # but makes type checking happy
        return to_core_table(entity.local_table)

    if callable(entity):
        return to_core_table(entity())

    raise NotImplementedError("to_core_table not implemented for entity type")



#!/usr/bin/env python

"""Setup script for the package."""

import logging
import os
import sys
from typing import Optional

import setuptools

PACKAGE_NAME = "sqla_ext"
MINIMUM_PYTHON_VERSION = "3.7"


def check_python_version() -> None:
    """Exit when the Python version is too low."""
    if sys.version < MINIMUM_PYTHON_VERSION:
        sys.exit("Python {0}+ is required.".format(MINIMUM_PYTHON_VERSION))


def read_package_variable(key: str, filename: str = "__init__.py") -> "Optional[str]":
    """Read the value of a variable from the package without importing."""
    module_path = os.path.join("src", PACKAGE_NAME, filename)
    with open(module_path) as module:
        for line in module:
            parts = line.strip().split(" ", 2)
            if parts[:-1] == [key, "="]:
                return parts[-1].strip("'").strip('"')
    logging.warning("'%s' not found in '%s'", key, module_path)
    return None


check_python_version()


DEV_REQUIRES = [
    "pytest",
    "pytest-cov",
    "pytest-benchmark",
    "pytest-asyncio",
    "pre-commit",
    "pylint",
    "black",
    "mypy",
    "asyncpg",
]

setuptools.setup(
    name=read_package_variable("__project__"),
    version=read_package_variable("__version__"),
    description="SQLAlchemy Helpers",
    url="https://github.com/olirice/sqla",
    author="Oliver Rice",
    author_email="oliver@oliverrice.com",
    packages=setuptools.find_packages("src", exclude=("src/tests",)),
    package_dir={"": "src"},
    tests_require=["pytest", "coverage"],
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=["typing_extensions", "sqlalchemy[mypy]>=1.4.0"],
    extras_require={
        "dev": DEV_REQUIRES,
        "docs": ["mkdocs", "mkautodoc", "pygments", "pymdown-extensions"],
    },
)



from __future__ import annotations

from typing import Callable, Union

from sqlalchemy import Table
from sqlalchemy.orm import Mapper
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class ORMTableProtocol(Protocol):
    __table__: Table


TableCoercable = Union[
    Table, ORMTableProtocol, Mapper, Callable[[], Union[ORMTableProtocol, Table]]
]



from typing import Any, Dict

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.functions import FunctionElement
from sqlalchemy.types import JSON


class build_object(FunctionElement):  # type: ignore
    r"""JSON object creation

    :Dialects:
        - mysql
        - postgresql
        - sqlite

    :param expression: A SQL expression, such as a
        :class:`ColumnElement` expression or a :class:`TextClause` with elements
        that are castable as JSON

    :return: :class:`FuntionElement`

    E.g.::

        from sqlalchemy import select
        from sqla_ext import func as func_ext

        query = select([
            func_ext.json.build_object("first_key", "first_value", "second_key", 2)
        ])

    The above statement will produce SQL resembling::

        SELECT
            jsonb_build_object('first_key', 'first_value', 'second_key', 2)
    """

    type = JSON()
    name = "jsonb_build_object"


@compiles(build_object, "postgresql")
def pg_build_object(
    element: build_object, compiler: SQLCompiler, **kw: Dict[str, Any]
) -> str:
    return "jsonb_build_object(%s)" % compiler.process(element.clauses, **kw)


@compiles(build_object, "sqlite")
@compiles(build_object, "mysql")
def other_build_object(
    element: build_object, compiler: SQLCompiler, **kw: Dict[str, Any]
) -> str:
    return "json_object(%s)" % compiler.process(element.clauses, **kw)
